/*
 * This file is part of Jenirok.
 *
 * Jenirok is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Jenirok is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Jenirok.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <QDebug>
#include <QApplication>
#include <QtCore/QTimerEvent>
#include <QtCore/QVariant>
#include <QtDBus/QDBusArgument>
#include <QtDBus/QDBusConnection>
#include <icd/dbus_api.h>
#include "connectionmanager.h"


ConnectionManager::ConnectionManager(QObject* parent): QObject(parent),
blocking_(true), stateReady_(false), connectionReady_(false), scanReady_(false),
connected_(false), timeout_(false), numberOfConnections_(0),
scannedConnections_(0), timer_(0), error_(NO_ERROR), connections_(0)
{
    QDBusConnection systemBus = QDBusConnection::systemBus();

    icd2interface_ = new QDBusInterface(ICD_DBUS_API_INTERFACE,
                                        ICD_DBUS_API_PATH, ICD_DBUS_API_INTERFACE,
                                        systemBus, this);

    systemBus.connect(ICD_DBUS_API_INTERFACE, ICD_DBUS_API_PATH,
                      ICD_DBUS_API_INTERFACE, ICD_DBUS_API_STATE_SIG,
                      this, SLOT(stateChange(const QDBusMessage&)));

    systemBus.connect(ICD_DBUS_API_INTERFACE, ICD_DBUS_API_PATH,
                      ICD_DBUS_API_INTERFACE, ICD_DBUS_API_CONNECT_SIG,
                      this, SLOT(connectionChange(const QDBusMessage&)));

    systemBus.connect(ICD_DBUS_API_INTERFACE, ICD_DBUS_API_PATH,
                      ICD_DBUS_API_INTERFACE, ICD_DBUS_API_SCAN_SIG,
                      this, SLOT(scanResult(const QDBusMessage&)));

}

ConnectionManager::~ConnectionManager()
{
    QDBusConnection systemBus = QDBusConnection::systemBus();

    systemBus.disconnect(ICD_DBUS_API_INTERFACE, ICD_DBUS_API_PATH,
                         ICD_DBUS_API_INTERFACE, ICD_DBUS_API_STATE_SIG,
                         this, SLOT(stateChange(const QDBusMessage&)));

    systemBus.disconnect(ICD_DBUS_API_INTERFACE, ICD_DBUS_API_PATH,
                         ICD_DBUS_API_INTERFACE, ICD_DBUS_API_CONNECT_SIG,
                         this, SLOT(connectionChange(const QDBusMessage&)));

    systemBus.disconnect(ICD_DBUS_API_INTERFACE, ICD_DBUS_API_PATH,
                         ICD_DBUS_API_INTERFACE, ICD_DBUS_API_SCAN_SIG,
                         this, SLOT(scanResult(const QDBusMessage&)));
}

void ConnectionManager::setBlocking(bool value)
{
    blocking_ = value;
}

bool ConnectionManager::connect()
{
    connectionReady_ = false;
    unsigned int flags = static_cast<unsigned int>(ICD_CONNECTION_FLAG_USER_EVENT);
    icd2interface_->call(ICD_DBUS_API_CONNECT_REQ, QVariant(flags));

    if(blocking_)
    {
        waitSignal(&connectionReady_);
        return connected_;
    }

    return true;
}

bool ConnectionManager::connect(ConnectionManager::Connection const& connection)
{
    return connect(connection.id);
}

bool ConnectionManager::connect(QString const& id)
{
    QDBusMessage msg = QDBusMessage::createMethodCall("com.nokia.icd",
                                                      "/com/nokia/icd",
                                                      "com.nokia.icd",
                                                      "connect");
    QList<QVariant> arguments;

    arguments.append(QVariant(id));

    unsigned int val = 0;

    arguments.append(QVariant(val));

    msg.setArguments(arguments);

    QDBusMessage rep = QDBusConnection::systemBus().call(msg);

    if(rep.type() == QDBusMessage::ErrorMessage)
    {
        if(rep.errorName() == "com.nokia.icd.error.invalid_iap")
        {
            error_ = INVALID_IAP;
        }
        else
        {
            error_ = UNKNOWN_ERROR;
        }

        return false;
    }

    return true;
}

bool ConnectionManager::getBestConnection(Connection& connection)
{
    bool blockingValue = blocking_;

    blocking_ = true;

    QList<Connection> connections;

    if(!scanConnections(connections))
    {
        return false;
    }

    blocking_ = blockingValue;

    if(connections.size() == 0)
    {
        qDebug() << "No connections";
        error_ = NO_AVAILABLE_CONNECTIONS;
        return false;
    }

    int biggestWlan = -1;
    int biggestGprs = -1;
    int bestWlan = -1;
    int bestGprs = -1;

    for(int i = 0; i < connections.size(); i++)
    {
       switch(connections.at(i).type)
       {
       case WLAN:
           if(connections.at(i).strength > biggestWlan)
           {
               biggestWlan = connections.at(i).strength;
               bestWlan = i;
           }
           break;

       case GPRS:
           if(connections.at(i).strength > biggestGprs)
           {
               biggestGprs = connections.at(i).strength;
               bestGprs = i;
           }
           break;

       default:
           qDebug() << "Unknown connection type";
       }
    }

    if(bestWlan >= 0)
    {
        connection = connections.at(bestWlan);
        return true;
    }
    else if(bestGprs >= 0)
    {
        connection = connections.at(bestGprs);
        return true;
    }
    else
    {
        error_ = NO_AVAILABLE_CONNECTIONS;
        return false;
    }

}

bool ConnectionManager::autoConnect()
{
   Connection best;

   if(!getBestConnection(best))
   {
       return false;
   }

   qDebug() << "Best connection: " << best.name;

   return connect(best.id);

}

bool ConnectionManager::disconnect(bool force)
{
    // Forced disconnect is not allowed if connection
    // was not initialized by this class
    if(!connected_ && force)
    {
        return false;
    }

    connectionReady_ = false;
    unsigned int flags;

    if(force)
    {
        flags = static_cast<unsigned int>(ICD_CONNECTION_FLAG_UI_EVENT);
    }
    else
    {
        flags = static_cast<unsigned int>(ICD_CONNECTION_FLAG_USER_EVENT);
    }

    icd2interface_->call(ICD_DBUS_API_DISCONNECT_REQ, QVariant(flags));
    connected_ = false;
    return true;
}

bool ConnectionManager::isConnected()
{
    stateReady_ = false;
    QDBusMessage rep = icd2interface_->call(ICD_DBUS_API_STATE_REQ);

    unsigned int numOfReplies = rep.arguments().value(0).value<unsigned int>();

    if(numOfReplies == 0)
    {
        emit isConnectedReply(false);
        return false;
    }

    if(blocking_)
    {
        waitSignal(&stateReady_);
        return connected_;
    }

    return true;
}

bool ConnectionManager::scanConnections(QList<ConnectionManager::Connection>& connections)
{
    unsigned int flags = static_cast<unsigned int>(ICD_SCAN_REQUEST_ACTIVE_SAVED);
    scanReady_ = false;
    scannedConnections_ = 0;
    connections_ = &connections;
    QDBusMessage rep = icd2interface_->call(ICD_DBUS_API_SCAN_REQ, QVariant(flags));

    numberOfConnections_ = rep.arguments().value(0).toList().size();

    if(numberOfConnections_ == 0)
    {
        connections_ = 0;
        return false;
    }

    if(blocking_)
    {
        bool ret = waitSignal(&scanReady_);
        connections_ = 0;
        return ret;
    }

    return true;
}

ConnectionManager::NetworkMode ConnectionManager::getNetworkMode()
{
    QDBusMessage msg = QDBusMessage::createMethodCall("com.nokia.phone.net",
                                                      "/com/nokia/phone/net",
                                                      "Phone.Net",
                                                      "get_registration_status");

    QDBusMessage rep = QDBusConnection::systemBus().call(msg);

    if(rep.type() == QDBusMessage::ErrorMessage)
    {
        qDebug() << rep.errorMessage();

        return NETWORK_UNKNOWN;
    }

    char services = rep.arguments().value(6).toChar().toAscii();

    qDebug() << services;

    if(services & 0x08)
    {
        return NETWORK_3_5G;
    }
    else if(services & 0x04)
    {
        return NETWORK_3G;
    }
    else if(services & 0x01)
    {
        return NETWORK_2_5G;
    }
    else if(services & 0x02)
    {
        return NETWORK_2G;
    }

    return NETWORK_UNKNOWN;
}

ConnectionManager::Error ConnectionManager::error() const
{
    return error_;
}

void ConnectionManager::stateChange(const QDBusMessage& rep)
{
    unsigned int status = rep.arguments().value(7).value<unsigned int>();

    switch(status)
    {
    case ICD_STATE_CONNECTING:
        break;
    case ICD_STATE_CONNECTED:
        connected_ = true;
        stateReady_ = true;
        break;
    case ICD_STATE_DISCONNECTING:
        break;
    case ICD_STATE_DISCONNECTED:
        connected_ = false;
        stateReady_ = true;
        break;
    case ICD_STATE_LIMITED_CONN_ENABLED:
        connected_ = true;
        stateReady_ = true;
        break;
    case ICD_STATE_LIMITED_CONN_DISABLED:
        connected_ = false;
        stateReady_ = true;
        break;
    case ICD_STATE_SEARCH_START:
        break;
    case ICD_STATE_SEARCH_STOP:
        break;
    case ICD_STATE_INTERNAL_ADDRESS_ACQUIRED:
        break;
    default:
        qDebug() << "Unknown connection status";
        break;
    }

    if(stateReady_)
    {
        emit isConnectedReply(connected_);
    }

}

void ConnectionManager::connectionChange(const QDBusMessage& rep)
{
    unsigned int status = rep.arguments().value(6).value<unsigned int>();

    switch(status)
    {
    case ICD_CONNECTION_SUCCESSFUL:
        connected_ = true;
        connectionReady_ = true;
        break;
    case ICD_CONNECTION_NOT_CONNECTED:
        connected_ = false;
        connectionReady_ = true;
        break;
    case ICD_CONNECTION_DISCONNECTED:
        connected_ = false;
        connectionReady_ = true;
        break;
    default:
        qDebug() << "Unknown connection status";
        break;
    }

    if(connectionReady_)
    {
        emit connectReply(connected_);
    }
}

void ConnectionManager::scanResult(const QDBusMessage& rep)
{
    if(!connections_)
    {
        return;
    }

    QList<QVariant> args = rep.arguments();

    unsigned int status = args.value(0).value<unsigned int>();

    if(status == ICD_SCAN_COMPLETE)
    {
        scannedConnections_++;
    }

    if(scannedConnections_ >= numberOfConnections_)
    {
        scanReady_ = true;
        connections_ = 0;
        emit scanReady();
        return;
    }

    if(status != ICD_SCAN_NEW)
    {
        return;
    }

    Connection connection;
    connection.id = QString(args.value(10).toByteArray());
    connection.name = args.value(8).toString();
    connection.strength = args.value(11).toInt();

    QString type = args.value(7).toString();

    if(type == "GPRS")
    {
        connection.type = GPRS;
    }
    else if(type == "WLAN_INFRA" || type == "WLAN_ADHOC")
    {
        connection.type = WLAN;
    }
    else
    {
        qDebug() << "Unknown connection type: " << type;
        return;
    }

    emit newConnection(connection);

    connections_->push_back(connection);
}

bool ConnectionManager::waitSignal(bool* ready)
{
    timeout_ = false;
    timer_ = startTimer(TIMEOUT);

    while(!*ready && !timeout_)
    {
        QCoreApplication::processEvents(QEventLoop::WaitForMoreEvents);
    }

    killTimer(timer_);

    return *ready || !timeout_;
}

void ConnectionManager::timerEvent(QTimerEvent* event)
{
    Q_UNUSED(event);
    killTimer(timer_);
    timeout_ = true;
    timer_ = 0;

    qDebug() << "Connection request timed out";
}
